class StringStream:
    """
    Progressively reads from a string whose length is monotonically increasing.

    The `reader` argument should be a nullary function that returns a string.
    """

    def __init__(self, reader):
        self.reader = reader
        self.offset = 0

    def read(self):
        output = self.reader()
        next_offset = len(output)

        if next_offset == self.offset:
            # No new data
            return None

        assert next_offset > self.offset
        fragment = output[self.offset:]
        self.offset = next_offset
        return fragment

    @classmethod
    def from_task(cls, task):
        return cls(lambda: task.get_output())



import asyncio
import logging
import signal

from ptyprocess import PtyProcess

from .color import ColorText
from .task import BufferedTask, TaskStatus, TaskAction, TaskEvent


class Process(BufferedTask):
    """
    A task that executes a process within a pseudo-terminal.

    * [argv]
      All positional arguments to the constructor are interpreted as the arguments to the process.
    * shell
      Not implemented yet.
    * cwd
      Optional current working directory for the process.
    * env
      Optional environment dictionary.
    * name
      Optional name for this task. If not provided, the command string is used as the name.
    """

    BLOCK_SIZE = 1024

    def __init__(self, *argv, shell=False, cwd=None, env=None, name=None):
        self.argv = list(map(str, argv))
        super().__init__(name=(name or ' '.join(self.argv)))
        # TODO(saumitro): Implement shell support
        self.shell = shell
        self.process = None
        self.cwd = cwd
        self.env = env
        self.future_process_exit = None
        self.future_restart = None

    def start(self):
        assert (self.process is None) or (not self.process.isalive())

        # Make sure any previously registered readers are first removed so that
        # they aren't accidentally invoked for the new process
        if self.process is not None:
            self.loop.remove_reader(self.process.fd)

        # Start the process
        self.process = PtyProcess.spawn(
            self.argv,
            cwd=self.cwd,
            env=self.env,
        )

        # Monitor process output
        self.loop.add_reader(self.process.fd, self.on_data_available)

        # Monitor process exit
        async def monitor_process_status(process):
            while process.isalive():
                await asyncio.sleep(0.5)
            self.on_process_terminate(process)
            self.future_process_exit = None
        self.future_process_exit = asyncio.ensure_future(monitor_process_status(self.process))

        # Transition status
        self.events.publish(TaskEvent.STATUS_CHANGED)

    def stop(self):
        # The force flag is treated as a last resort. PtyProcess first attempts
        # to amicably terminate the process with SIGHUP/SIGINT.
        self.process.terminate(force=True)

    def restart(self):
        if self.future_restart is not None:
            return

        async def stop_then_start():
            if self.future_process_exit is not None:
                self.stop()
                await self.future_process_exit
            self.start()
            self.future_restart = None

        self.future_restart = asyncio.ensure_future(stop_then_start())

    def on_data_available(self):
        if not self.process.isalive():
            return
        try:
            self.write_output(self.process.read(self.BLOCK_SIZE).decode('utf8'))
        except EOFError:
            pass

    def on_process_terminate(self, process):
        if process != self.process:
            logging.warn(f'Internal inconsistency: process changed before termination callback.')
            return
        self.loop.remove_reader(self.process.fd)
        self.output_termination_message(self.get_status())
        self.events.publish(TaskEvent.STATUS_CHANGED)

    def output_termination_message(self, status):
        # TODO(saumitro): Also output human readable uptime
        colorant = ColorText.red
        reason = 'due to unknown reasons'
        if self.process.exitstatus is not None:
            colorant = ColorText.red if status == TaskStatus.FAILED else ColorText.green
            reason = f'with code {self.process.exitstatus}'
        elif self.process.signalstatus is not None:
            signal_name = signal.Signals(self.process.signalstatus).name
            reason = f'due to signal {self.process.signalstatus} ({signal_name})'
        self.write_output(colorant(f'\n[Process exited {reason}]\n\n'))

    def get_actions(self):
        return [
            TaskAction(name='Stop', handler=self.stop),
            TaskAction(name='Restart', handler=self.restart)
        ]

    def get_status(self):
        if self.process.isalive():
            return TaskStatus.ACTIVE
        if self.process.exitstatus == 0:
            return TaskStatus.FINISHED
        return TaskStatus.FAILED



from pathlib import Path

from .process import Process
from .watch import Watch


def auto_rsync(
    source,
    dest,
    *,
    name,
    verbose=True,
    archive=True,
    recursive=True,
    # If true, skip files that are newer on the receiver
    update=False,
    # If true, delete extraneous files from dest dirs
    delete=False,
    compress=True
):
    command = ['rsync', source, dest]
    flag_map = {
        'archive': archive,
        'compress': compress,
        'delete': delete,
        'recursive': recursive,
        'update': update,
        'verbose': verbose,
    }
    for flag, is_set in flag_map.items():
        if is_set:
            command.append('--' + flag)

    # Establish the directory to watch for changes
    watch_dir = Path(source)
    if not watch_dir.is_dir():
        # Source is a file. Watch its parent directory.
        watch_dir = watch_dir.parent

    return Watch(
        # Watch the source path for changes
        path=watch_dir,
        # Sync the destination using rsync when changes are detected
        handler=Process(*command, name='rsync'),
        # Ensure the destination is synced at start
        force_once=True,
        recursive=recursive,
        name=name
    )



import enum
import io
import itertools

from .event import EventEmitter


class TaskStatus(enum.Enum):
    ACTIVE = 'active'
    FINISHED = 'finished'
    FAILED = 'failed'


class TaskEvent(enum.Enum):
    OUTPUT_UPDATED = 'output-updated'
    STATUS_CHANGED = 'status-changed'


class TaskAction:
    def __init__(self, name, handler):
        self.name = name
        self.handler = handler


class Task:
    """
    Abstract base class for tasks: an asynchronous worker.
    """

    id_generator = itertools.count()

    def __init__(self, name):
        self.id = next(Task.id_generator)
        self.name = name
        self.events = EventEmitter()
        self.loop = None

    def start(self):
        """
        Register tasks with the event loop
        """
        raise NotImplementedError

    def stop(self):
        """
        Stop the task and remove from the event loop
        """
        raise NotImplemented

    def get_output(self):
        """
        Provide the output for this task
        """
        raise NotImplementedError

    def get_actions(self):
        """
        Returns a list of strings describing the available actions for this task.
        """
        return []

    def get_status(self):
        """
        Returns a value from TaskStatus.
        """
        raise NotImplementedError

    def perform_action(self, action_name):
        """
        Execute the given action, where action is one of the strings returned
        the get_actions method.
        """
        for action in self.get_actions():
            if action.name == action_name:
                return action.handler()
        raise ValueError('Unsupported action: {}'.format(action_name))

    def start_subtask(self, task):
        """
        Initiate a task owned by `self` rather than the session.
        """
        assert self.loop is not None
        task.loop = self.loop
        task.start()

    @property
    def is_active(self):
        return self.get_status() == TaskStatus.ACTIVE


class BufferedTask(Task):
    """
    A task that stores its output in an in-memory buffer.
    """

    def __init__(self, name):
        super().__init__(name)
        self.output = io.StringIO()

    def write_output(self, text):
        self.output.write(text)
        self.events.publish(TaskEvent.OUTPUT_UPDATED)

    def write_line(self, line):
        self.write_output(line + '\n')

    def get_output(self):
        return self.output.getvalue()



def colorant(color_code):
    """
    Returns a function that formats strings with the given ANSI color code.
    """
    prefix = '\033[' + str(color_code) + 'm'

    def color_text(text):
        return prefix + text + '\033[0m'

    return color_text


class ColorText:
    """
    A collection of ANSI color code formatters.
    """
    blue = colorant(94)
    green = colorant(92)
    magenta = colorant(95)
    red = colorant(91)
    yellow = colorant(93)


def status_ok(tag, msg):
    print(f'{ColorText.green(tag) + ":":20} {msg}')



import functools
import watchdog.observers

from .color import ColorText
from .stream import StringStream
from .task import Task, BufferedTask, TaskEvent, TaskStatus, TaskAction


class Watch(BufferedTask):
    """
    Watches a path and invokes a task whenever a change is detected.

    * path
      The path to watch for changes.
    * handler
      Either a task instance that can be started multiple times, or a function that
      that accepts a list of watchdog fs events and returns a task to execute.
    * recursive
      If true, the given path is recursively watched.
    * delay
      A debounce interval for accumulating changes (in seconds).
    * force_once
      If true, the handler is invoked unconditionally when the watch starts.
    """

    def __init__(self, path, *, handler, name, recursive=False, delay=1.0, force_once=False):
        super().__init__(name)
        self.path = str(path)
        self.handler = (lambda _: handler) if isinstance(handler, Task) else handler
        self.delay = delay
        self.force_once = force_once
        self.queue = []
        self.pending_consume = False
        self.active_task = None
        self.active_task_stream = None
        self.observer = watchdog.observers.Observer()
        self.observer.schedule(self, self.path, recursive=recursive)

    def start(self):
        self.write_line(f'{ColorText.green("Watching:")} {self.path}')
        self.observer.start()
        if self.force_once:
            self.write_line('Invoking task once unconditionally')
            self.initiate_consume()

    def stop(self):
        if self.active_task is not None:
            self.active_task.stop()
        self.observer.stop()
        self.observer.join()

    def get_status(self):
        return TaskStatus.ACTIVE

    def dispatch(self, event):
        self.queue.append(event)
        self.initiate_consume()

    def initiate_consume(self, delay=None):
        if self.pending_consume:
            return
        self.pending_consume = True
        if delay is None:
            delay = self.delay
        self.loop.call_soon_threadsafe(self.loop.call_later, delay, self.consume)

    def consume(self):
        if self.active_task is not None:
            # Wait for the task to complete. Events will keep queueing until then.
            self.loop.call_later(self.delay, self.consume)
            return
        events, self.queue = self.queue, []
        self.active_task = self.handler(events)
        self.active_task_stream = StringStream.from_task(self.active_task)
        self.active_task.events.subscribe(self.on_task_event)
        self.start_subtask(self.active_task)
        self.pending_consume = False

    def on_task_event(self, event):
        if event == TaskEvent.OUTPUT_UPDATED:
            self.write_output(self.active_task_stream.read())
        elif event == TaskEvent.STATUS_CHANGED:
            if not self.active_task.is_active:
                self.active_task.events.unsubscribe(self.on_task_event)
                self.active_task = None
                self.active_task_stream = None

    def get_actions(self):
        return [
            TaskAction(
                name='Trigger',
                handler=functools.partial(self.initiate_consume, delay=0)
            )
        ]



import errno
import logging
import socket

from pathlib import Path

import tornado
import tornado.log
import tornado.platform.asyncio
import tornado.web
import tornado.websocket

from .color import status_ok
from .stream import StringStream
from .task import TaskEvent


class DashboardHandler(tornado.web.RequestHandler):
    def initialize(self, session):
        self.session = session

    def get(self):
        self.render('app.html', session=self.session)


class TaskInfoHandler(tornado.websocket.WebSocketHandler):
    def initialize(self, session):
        self.session = session

    def open(self):
        self.session.task_status_monitor.subscribe(self.on_task_status_update)
        self.send_task_info()

    def on_task_status_update(self, _):
        self.send_task_info()

    def send_task_info(self):
        self.write_message({'tasks': [self.get_task_info(task) for task in self.session.tasks]})

    def get_task_info(self, task):
        return {
            'id': task.id,
            'name': task.name,
            'status': task.get_status().value,
            'actions': [action.name for action in task.get_actions()],
        }

    def on_close(self):
        self.session.task_status_monitor.unsubscribe(self.on_task_status_update)


class TaskOutputHandler(tornado.websocket.WebSocketHandler):
    def initialize(self, session):
        self.session = session
        self.task = None
        self.stream = None

    def open(self, task_id):
        self.task = self.session.get_task_by_id(int(task_id))
        self.stream = StringStream.from_task(self.task)
        self.task.events.subscribe(self.on_task_event)
        self.send_output_to_client()

    def on_close(self):
        self.unsubscribe()

    def on_task_event(self, event):
        if event == TaskEvent.OUTPUT_UPDATED:
            self.send_output_to_client()

    def send_output_to_client(self):
        new_fragment = self.stream.read()
        if new_fragment is None:
            return
        try:
            self.write_message(new_fragment)
        except tornado.websocket.WebSocketClosedError:
            self.unsubscribe()

    def unsubscribe(self):
        if self.task is not None:
            self.task.events.unsubscribe(self.on_task_event)


class TaskActionHandler(tornado.web.RequestHandler):
    def initialize(self, session):
        self.session = session

    def post(self):
        payload = tornado.escape.json_decode(self.request.body)
        task = self.session.get_task_by_id(payload['task'])
        result = task.perform_action(payload['action'])
        self.write(result or 'OK')


class WebApp(tornado.web.Application):
    def __init__(self, session):
        static_path = Path(__file__).parent / 'client' / 'static'
        template_path = static_path / 'templates'
        session_dict = {'session': session}

        handlers = [
            (
                r'/',
                DashboardHandler,
                session_dict
            ),
            (
                r'/tasks',
                TaskInfoHandler,
                session_dict
            ),
            (
                r'/task-output/(.*)',
                TaskOutputHandler,
                session_dict
            ),
            (
                r'/task-action',
                TaskActionHandler,
                session_dict
            ),
            (
                r'/static/(.*)',
                tornado.web.StaticFileHandler,
                {'path': str(static_path)}
            ),
        ]

        super().__init__(handlers, template_path=str(template_path))

    def get_port_candidates(self, preferred_port, max_distance=10):
        for dist in range(max_distance):
            yield preferred_port + dist
            if 0 > dist > preferred_port:
                yield preferred_port - dist

    def listen(self, port, address):
        for port in self.get_port_candidates(preferred_port=port):
            try:
                server = super().listen(port=port, address=address)
                break
            except socket.error as err:
                if err.errno == errno.EADDRINUSE:
                    continue
                raise
        status_ok('Listening', f'http://{address or "localhost"}:{port}/')
        return server

    @classmethod
    def setup(cls):
        # Install the Tornado I/O loop singleton
        tornado.platform.asyncio.AsyncIOMainLoop().install()
        tornado.log.enable_pretty_logging()
        logging.getLogger().setLevel(logging.WARNING)



class EventEmitter:
    """
    A simple pub-sub helper.
    """

    def __init__(self):
        self.subscribers = set()

    def subscribe(self, subscriber):
        self.subscribers.add(subscriber)

    def unsubscribe(self, subscriber):
        self.subscribers.remove(subscriber)

    def publish(self, event):
        # Clone the current subscribers since a subscriber
        # might mutate the subscription in response to a published event.
        for subscriber in set(self.subscribers):
            subscriber(event)



import os
import pathlib

# This acquires the concrete subclass for this platform (eg: PosixPath)
ConcretePath = type(pathlib.Path())


class Path(ConcretePath):
    """
    A pathlib.Path subclass that attempts to preserve the trailing
    path separator (which is stripped by the stdlib implementation).

    This is useful for various posix utilities like rsync where a trailing
    slash in the path can be significant.
    """

    def __new__(cls, *args, **kwargs):
        path = super().__new__(cls, *args, **kwargs)
        if args:
            path._capture_trailing_separator(args[-1])
        return path

    def _capture_trailing_separator(self, last_part):
        if last_part.endswith(os.sep):
            self._str_override = str(self) + os.sep
        return self

    def __truediv__(self, key):
        return super().__truediv__(key)._capture_trailing_separator(key)

    def __str__(self):
        return getattr(self, '_str_override', super().__str__())



from setuptools import setup, find_packages


def get_long_description():
    with open("README.md", "r") as readme_file:
        return readme_file.read()


setup(
    name='oxen',
    version='0.1.0',
    description='Task runner with a web-based frontend',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author='Saumitro Dasgupta',
    author_email='sd@cs.stanford.edu',
    url='https://github.com/ethereon/oxen',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
    ],
    packages=find_packages(),
    package_data={
        'oxen': [
            'client/static/js/*.js*',
            'client/static/css/*.css',
            'client/static/templates/*.html',
        ],
    },
    include_package_data=True,
    install_requires=['tornado', 'watchdog'],
    zip_safe=True,
    license='BSD'
)



import asyncio
import functools
import pathlib
import sys

from collections import OrderedDict

from .color import status_ok
from .event import EventEmitter
from .task import TaskEvent
from .webserver import WebApp


class TaskStatusMonitor(EventEmitter):
    """
    Re-publishes status changes from a collection of tasks.
    """

    def add(self, task):
        task.events.subscribe(functools.partial(self.on_task_event, task=task))

    def on_task_event(self, event, *, task):
        if event == TaskEvent.STATUS_CHANGED:
            self.publish(task)


class Session:
    """
    Manages the execution and monitoring of a collection of tasks.
    """

    def __init__(self, name=None):
        self.name = name
        self.id_to_tasks = OrderedDict()
        self.webserver = WebApp(session=self)
        self.task_status_monitor = TaskStatusMonitor()

    def add_task(self, task):
        self.id_to_tasks[task.id] = task

    def __iadd__(self, task):
        self.add_task(task)
        return self

    def start_tasks(self, loop):
        for task in self.tasks:
            task.loop = loop
            task.start()
            self.task_status_monitor.add(task)
            status_ok('Started', task.name)

    def get_task_by_id(self, task_id):
        return self.id_to_tasks[task_id]

    def start(self, *, port=4242, address=''):
        WebApp.setup()
        loop = asyncio.get_event_loop()
        # Register all tasks
        self.start_tasks(loop)
        # Setup webserver
        self.webserver.listen(port=port, address=address)
        loop.run_forever()

    @property
    def tasks(self):
        return self.id_to_tasks.values()
